title Set Environment Variable Program.
page 60,120			; page length = 60
				; page width  = 120
;
;	This program places a variable in the environment.
;
;	SETENV : Copyright (c) 1988 by John Wolchak.
;	  I give permission to alter the code, but not to copy or 
;	  redistribute it without my explicit permission.  If you 
;	  alter the code, please document changes and send me a 
;	  copy, so all can have it.  This document must always 
;	  accompany the SETENV program when a copy is made for 
;	  another machine.  This code, to the best of my 
;	  knowledge works well.  I disclaim any responsibility 
;	  for the codes actions (use at your own risk).  
;
;	John Wolchak.
;	56 Physics Building
;	Saskatoon, Saskatchewan
;	Canada     S7K 0W0
;	University of Saskatchewan.
;	Phone: (306) 966-4852
;	NetNorth (BITNET): Wolchak@sask
;
;	To assemble:
;	masm setenv,,,
;	link setenv,,,
;	exe2bin setenv.exe setenv.com
;
code segment
assume cs:code,ds:code
org 100h

start:		jmp beginning
;
;		Data Area
;
exit_code	db	03h		; Environment variable created.
					; PSP is Program Segment Prefix.
cpsp		dw	00h		; The psp of command.com.
env_off		dw	00h		; Environment offset address.
env_siz		dw	00h		; Environment size in bytes.
s_data		dw	00h		; Size of environment data.
a_env_var	dw	00h		; Address of environment variable.
s_env_var	dw	00h		; Environment variable size.
d_env_var	db	80 dup(0)
null		db	00h,00h		; Environment tail.
y_prompt	dw	08h		; Size of '%prompt '.
x_prompt	dw	06h		; Size.
l_prompt	db	"prompt"
u_prompt	db	"PROMPT"
o_prompt	dw	00h		; Offset of %prompt.
s_def		dw	04h		; Size.
l_def		db	"%def"
u_def		db	"%DEF"
x_def		dw	00h		; Default value size.
v_def		dw	64 dup(0)	; Default value.
s_dosv		dw	04h		; Size.
l_dosv		db	"dosv"
u_dosv		db	"DOSV"
s_dosm		dw	04h		; Size.
l_dosm		db	"dosm"
u_dosm		db	"DOSM"
s_cwd		dw	03h		; Size.
l_cwd		db	"cwd"
u_cwd		db	"CWD"
s_drive		dw	05h		; Size.
l_drive		db	"drive"
u_drive		db	"DRIVE"
s_path		dw	05h		; Size.
d_path		db	"PATH="
s_comspec	dw	08h		; Size.
d_comspec	db	"COMSPEC="
s_prompt	dw	07h		; Size.
d_prompt	db	"PROMPT="
author		db	"Set Environment Variable Program V1.0 (c) 18-Nov-1988 "
		db	"by John Wolchak.",13,10
help		db	"Usage is:  SETENV <variable> <value>",13,10,9,9
		db	"<variable> by it's self to erase.",13,10,9,9
		db	"<value> can include keys such as:",13,10,9,9
		db	"  %dosv / %dosm for DOS major / minor version."
		db	13,10,9,9
		db	"  %cwd for current directory.",13,10,9,9
		db	"  %drive for default drive.",13,10,9,9
		db	"  %+n / %-n for default drive plus / minus 'n'."
		db	13,10,9,9
		db	"<value> can be '%prompt <message>' to ask for a value, or"
		db	13,10,9,9
		db	"  '%prompt %def<del><text><del> <message>' to ask for a value."
		db	13,10
		db	"Example:   setenv drivedir %drive:%cwd",13,10
		db	"Note:	   Use %% to represent a single % in BAT files."
		db	13,10,13,10
		db	"The program returns the following codes:",13,10,9
		db	"no command, help issued 		00h",13,10,9
		db	"environment variable erased		02h",13,10,9
		db	"environment variable created		03h",13,10,9
		db	"environment space is over written	04h",13,10,9
		db	"environment variable updated		05h",13,10,9
		db	"unmatched delimiter			06h",13,10,9
;		db	"environment variable not found		08h",13,10,9
		db	"offset of drive will be invalid 	09h",13,10,9
		db	"environment space is full		10h"
		db	"$"
;
;		Macro Area
;
compare_str	macro	num
		mov	cx, num
		rep	cmpsb
		endm

move_str	macro	num
		mov	cx, num
		rep	movsb
		endm

;
;		Code Area
;
Beginning:
		mov si, 80h		; Offset of command line.
		lodsb			; Get a byte.
		cmp al, 00h		; Is there any data.
		jnz cont		; Yes. Jump.
dis_help:
		mov dx, offset author	; Help message.
		mov ah, 09h		; Display it.
		int 21h
		mov exit_code, 00h	; No command, help issued.
		jmp nor_term		; Exit.

;		Get rid of leading spaces and tabs.
cont:
		lodsb			; Get a byte.
		cmp al, " "		; Is it leading blank.
		jz cont			; Yes.  Loop.
		cmp al, 09h		; Is it leading tab.
		jz cont			; Yes.  Loop.
		cmp al, 0dh		; Is it a carriage return.
		jz dis_help		; Yes. dis_help.
		dec si			; Backup for the character.

		mov a_env_var, si	; Offset of the variable.
		mov di, offset d_env_var ; Offset of env. variable.

;		Get the variable name,
;		and convert to upper case.
get_var:
		lodsb			; Get a byte.
		cmp al, " "		; Is it the end of the variable?
		jz var_end		; Yes.  Jump out.
		cmp al, 09h		; Is it the end of the variable?
		jz var_end		; Yes.  Jump out.
		cmp al, "="		; Is it the end of the variable?
		jz var_end		; Yes. Jump out.
		cmp al, 0dh		; Is it the end.
		jz no_data		; Yes. We got it.
		cmp al, "z"		; Is
		jg not_lower		;   it
		cmp al, "a"		;     lower
		jb not_lower		;       case?
		sub al, 20h		; Make it upper case.
not_lower:
		stosb			; Save the byte.
		jmp get_var

no_data:
		mov al, 0dh
		mov [si], al
var_end:
		mov al, "="		; Set the end of variable.
		stosb			; Save it.

		mov ax, si		; Get the offset.
		mov bx, a_env_var	; Beginning of the variable.
		sub ax, bx		; Get the variable lenght.
		mov s_env_var, ax	; And save it.

		mov al, [si]		; Get a byte.
		cmp al, "%"		; Is it a special one.
		jz prompt		; Yes. Jump out.

;		Get the data part.
get_data:
		mov al, [si]		; Get a byte.
		cmp al, "%"		; Is it a special one.
		jz percent_near		; Yes. Jump out.
get_next:
		lodsb			; Get a byte.
		cmp al, 0dh		; Is it the end.
		jz data_end_near	; Yes. We got it.
		stosb			; Save the byte.
		jmp get_data

data_end_near:
		jmp data_end
percent_near:
		jmp percent

;		Handle %prompt.
prompt:
		mov o_prompt, si
		push di
		push si

		mov di, si		; Set destination.
		inc di
		mov si, offset l_prompt	; Set Source.
		mov cx, x_prompt
		rep cmpsb		; Is it %prompt.
		jz default_near		; Yes.

		pop si
		push si
		mov di, si		; Set destination.
		inc di
		mov si, offset u_prompt	; Set source.
		mov cx, x_prompt
		rep cmpsb		; Is it %PROMPT.
		jz default_near		; Yes.

		pop si
		pop di
		jmp get_data		; No. Jump out.

default_near:
		jmp default
get_prompt:
		pop si
		push si

		mov ax, si		; Jump
		mov cx, y_prompt	;   the
		add ax, cx		;     %prompt
		mov si, ax		;       string.
		push si

prompt1:
		lodsb			; Scan for
		cmp al, 0dh		;   a return.
		jnz prompt1

		mov ah, "$"		; Insert string
		mov -1[si], ah		;   terminator.
		mov [si], al

		pop si
		mov dx, si		; Print
		mov ah, 09h		;   the prompt.
		int 21h

		pop si
		push si
		mov di, si
prompt2:
		mov ah, 01h		; Read keyboard and echo.
		int 21h

		cmp al, 08h		; Got a backspace character.
		jz prompt3
		cmp al, 7fh		; Got a delete character.
		jnz prompt4
		mov dl, 08h		; Print a backspace.
		mov ah, 02h
		int 21h
prompt3:
		dec di
		mov dl, " "		; Print a space.
		mov ah, 02h
		int 21h

		mov dl, 08h		; Print a backspace.
		mov ah, 02h
		int 21h
		jmp prompt2
prompt4:
		stosb			; Save the byte.
		cmp al, 0dh		; Got a return.
		jnz prompt2

		mov dl, 0dh		; Print a return.
		mov ah, 02h
		int 21h
		mov dl, 0ah		; Print a linefeed.
		mov ah, 02h
		int 21h

		pop si
		push si
		mov al, [si]
		cmp al ,0dh		; Is it a return.
		jnz prompt9		; No.  Jump out.

		mov di, si		; Move
		mov si, offset v_def	;   the default
		mov cx, x_def		;     in place.
		rep movsb

		mov al, 0dh
		stosb
prompt9:
		pop si
		pop di

		jmp get_data

;		Handle %def.
default:
		mov si, di
default0:
		lodsb
		cmp al, " "		; Is it space?
		jz default0		; Yes. Ignore.
		cmp al, 09h		; Is it a tab?
		jz default0		; Yes. Ignore.

		dec si
		push si
		mov di, si		; Set destination.
		mov si, offset l_def	; Set Source.
		mov cx, s_def
		rep cmpsb		; Is it %def.
		jz get_default		; Yes.

		pop si
		mov di, si		; Set destination.
		mov si, offset u_def	; Set source.
		mov cx, s_def
		rep cmpsb		; Is it %DEF.
		jz get_default		; Yes.

		jmp get_prompt

get_default:
		mov si, di
		lodsb			; Get the delimiter.
		mov dl, al		; Save it.
		mov di, offset v_def	; Get address for save of value.
		mov x_def, si		; Save the start of the value.
default1:
		lodsb			; Get a byte.
		cmp al, 0dh		; Is it a return?
		jz default2		; Yes. Jump out.

		cmp al, dl		; Is it the delimiter?
		jz default3		; Yes. Jump out.
		stosb			; Save it.
		jmp default1
default2:
		mov -1[si], dl		; No
		mov ah, " "		;   delimiter
		mov [si], ah		;     so
		mov ah, ">"		;       create
		mov 1[si], ah		;         the
		mov 2[si], al		;           info.
		dec si
		mov cs:[exit_code], 06h	; Unmatched delimter.
		jmp default1

default3:
		mov ax, si		; Get current pointer.
		mov bx, x_def		; Get default value start.
		sub ax, bx		; To get lenght of default value.
		dec ax
		mov x_def, ax		; Save it.

		mov ax, si		; Get cuurent pointer.
		mov bx, o_prompt	; Get prompt value start.
		sub ax, bx		; To get prompt noise length.
		inc ax
		mov y_prompt, ax	; Save it.

		pop si
		jmp get_prompt

get_data_near:
		jmp get_data
percent:
		mov al, 1[si]
		cmp al, "+"		; Is it a plus.
		jz off_drive

		cmp al, "-"		; Is it a minus.
		jnz cwd

;		Offset from a given drive.
off_drive:
		mov dh, al		; Save the sign.
		mov dl, 2[si]
		cmp dl, "0"		; Is
		jb get_data_near	;   it
		cmp dl, "9"		;     a
		jg get_data_near	;       number.
		sub dl, 30h		; Make it binary.

		mov ah, 19h		; Get the disk drive.
		int 21h
		add al, "A"		; Make it a letter.
		cmp dh, "+"
		jz add1
		sub al, dl		; Subtract from current drive.
		cmp al, "A"		; If less then A
		jge add2		;   we have an error.
let_error:
		mov exit_code, 09h	; Offset of drive will be invalid.
		mov al, "%"		; Put percent back.
		jmp get_next

add1:
		add al, dl		; Add from current drive.
		cmp al, "Z"		; If greater then Z
		jg let_error		;   we have an error.
add2:
		mov [si], al		; Save the letter.

;		Get rid of the %- or %+.
		push di
		push si
		inc si
		mov di, si		; Point at the - or +.
		mov ax, si		; Point
		add ax, 02h		;   after
		mov si, ax		;     the digit.
add9:
		lodsb			; Compress
		stosb			;   the
		cmp al, 0dh		;     data.
		jnz add9

		pop si
		pop di
		jmp get_data


;		Handle %cwd.
cwd:
		push di
		push si

		mov di, si		; Set destination.
		inc di
		mov si, offset l_cwd	; Set Source.
		mov cx, s_cwd
		rep cmpsb		; Is it %cwd.
		jz get_cwd		; Yes.

		pop si
		push si
		mov di, si		; Set destination.
		inc di
		mov si, offset u_cwd	; Set source.
		mov cx, s_cwd
		rep cmpsb		; Is it %CWD.
		jz get_cwd		; Yes.

		pop si
		pop di
		jnz dosv		; No. Jump out.

get_cwd:
		pop si
		pop di
		mov al, "\"		; Put in the
		mov [si], al		;   root backslash.
		push di
		push si
		mov si, offset help	; Place to put the directory name.

		mov ah, 47h		; Get current directory.
		mov dl, 00h		; Current disk drive.
		int 21h

cwd1:
		lodsb			; Find the
		cmp al, 00h		;   string
		jnz cwd1		;     terminator (zero).

		dec si
		mov di, si		; Directory name plus 1.
		pop si
		push si

		inc si
		inc si			; Advance
		inc si			;   over
		inc si			;     %cwd.
cwd2:
		lodsb			; Move the
		stosb			;   command string
		cmp al, 0dh		;     behind directory name.
		jnz cwd2

		pop si
		push si
		mov di, si		; Current point of command.
		mov si, offset help	; Start of directory.
		inc di
cwd3:
		lodsb			; Move the
		stosb			;   command string
		cmp al, 0dh		;     back.
		jnz cwd3

		pop si
		pop di

		jmp get_data

;		Handle %dosv.
dosv:
		push di
		push si

		mov di, si		; Set destination.
		inc di
		mov si, offset l_dosv	; Set Source.
		mov cx, s_dosv
		rep cmpsb		; Is it %dosv.
		jz get_dosv		; Yes.

		pop si
		push si
		mov di, si		; Set destination.
		inc di
		mov si, offset u_dosv	; Set source.
		mov cx, s_dosv
		rep cmpsb		; Is it %DOSV.
		jz get_dosv		; Yes.

		pop si
		pop di
		jmp dosm		; No. Jump out.

get_dosv:
		pop si
		pop di
		mov ah, 30h		; Get DOS version number.
		int 21h

		add al, "0"		; Make it a number.
		mov [si], al

;		Get rid of the "dosv" of the %dosv.
		push di
		push si
		inc si
		mov di, si		; Point at the "d".
		mov ax, si		; Point
		add ax, 04h		;   after
		mov si, ax		;     the "v".
dosv1:
		lodsb			; Compress
		stosb			;   the
		cmp al, 0dh		;     data.
		jnz dosv1

		pop si
		pop di
		jmp get_data

;		Handle %dosm.
dosm:
		push di
		push si

		mov di, si		; Set destination.
		inc di
		mov si, offset l_dosm	; Set Source.
		mov cx, s_dosm
		rep cmpsb		; Is it %dosm.
		jz get_dosm		; Yes.

		pop si
		push si
		mov di, si		; Set destination.
		inc di
		mov si, offset u_dosm	; Set source.
		mov cx, s_dosm
		rep cmpsb		; Is it %DOSM.
		jz get_dosm		; Yes.

		pop si
		pop di
		jmp drive		; No. Jump out.

get_dosm:
		pop si
		pop di
		mov ah, 30h		; Get DOS version number.
		int 21h

		mov cx, 04h
		cmp ah, 0ah		; Is it less then ten.
		jb dosm0		; Yes. Jump out.
		mov al, 00h
		mov dl, 0ah		; Break
		mov al, ah		;   up
		cbw			;     the
		div dl			;       digits.
		add al, "0"		; Make it a number.
		mov [si], al		; Store tens.
		mov cx, 03h
		inc si
dosm0:
		add ah, "0"		; Make it a number.
		mov [si], ah		; Store ones.

;		Get rid of the "dosm" of the %dosm.
		push di
		push si
		inc si
		mov di, si		; Point at the "d".
		mov ax, si		; Point
		add ax, cx		;   after
		mov si, ax		;     the "m".
dosm1:
		lodsb			; Compress
		stosb			;   the
		cmp al, 0dh		;     data.
		jnz dosm1

		pop si
		pop di
		cmp cx, 04h
		jz dosm2
		dec si
dosm2:
		jmp get_data

;		Handle %drive.
drive:
		push di
		push si

		mov di, si		; Set destination.
		inc di
		mov si, offset l_drive	; Set Source.
		mov cx, s_drive
		rep cmpsb		; Is it %drive.
		jz get_drv		; Yes.

		pop si
		push si
		mov di, si		; Set destination.
		inc di
		mov si, offset u_drive	; Set source.
		mov cx, s_drive
		rep cmpsb		; Is it %DRIVE.
		jz get_drv		; Yes.

		pop si
		pop di
		jmp get_next		; No. Jump out.

get_drv:
		pop si
		pop di
		mov ah, 19h		; Get the disk drive.
		int 21h

		add al, "A"		; Make it a letter.
		mov [si], al

;		Get rid of the "drive" of the %drive.
		push di
		push si
		inc si
		mov di, si		; Point at the "d".
		mov ax, si		; Point
		add ax, 05h		;   after
		mov si, ax		;     the "e".
drive1:
		lodsb			; Compress
		stosb			;   the
		cmp al, 0dh		;     data.
		jnz drive1

		pop si
		pop di
		jmp get_data

data_end:
		dec si			; We are N+1.
		mov ax, si		; Get the offset.
		mov bx, a_env_var	; The beginng of the variable.
		sub ax, bx		; Get the data lenght.
		add ax, 02h		; For two zero.
		mov s_data, ax		; And save it.

get_psp:
		push ds			; Save the segment base.

;		Critical Error Address Segment of program.
		mov si, 14h
		mov ax, [si]

;		Critical Error Address Segment of command.com.
		mov ds, ax
		mov si, 16h
		mov ax, [si]
		mov cs:[cpsp], ax

;		Segment where environment space.
		mov ds, ax
		mov si, 00h
get4d:
;		Look for 4dh on segment boundary.
		mov al, [si]
		cmp al, 4dh
		jz got4d

inc_seg:
		mov ax, si		; Increment
		add ax, 10h		;   to the next
		mov si, ax		;      segment boundary.
		jmp get4d

got4d:
		mov ax, 1[si]		; Get the psp.
		cmp cs:[cpsp], ax	; Is it the same as cpsp?
		jnz inc_seg		; No.  Jump out.

		mov ah, 00h
		mov al, 3[si]		; Get the number of segments.
		mov bx, 10h
		mul bx			; Get the number of bytes.
		mov cs:[env_siz], ax	; Save it.

		mov ax, si		; Increment
		add ax, 10h		;   to the environment
		mov si, ax		;      segment boundary.
		mov cs:[env_off], ax	; Save it.

;		Just a double check for environment space.
;		Check for PATH= or PROMPT= or COMPEC=.
		mov ax, ds		; Get the data segment
		mov es, ax		; and put it in extra segment.
		mov di, si		; This is our destination.

		pop ds			; Restore the data segment.
		mov si, offset d_path	; Check
		mov cx, s_path		;   for
		rep cmpsb		;     PATH=.
		jz got_env

		mov di, env_off		; Restore the destination.
		mov si, offset d_prompt	; Check
		mov cx, s_prompt	;   for
		rep cmpsb		;     PROMPT=.
		jz got_env

		mov di, env_off		; Restore the destination.
		mov si, offset d_comspec ;Check
		mov cx, s_comspec	;    for
		rep cmpsb		;      COMSPEC=.
		jz got_env

		push ds			; Save the data segment.
		mov ax, es		; Get the
		mov ds, ax		;   data segment
		mov si, cs:[env_off]	;     and source
		jmp get4d		;       to scan some more.

got_env:
		push ds			; Save the data segment.
		mov ax, es		; Get the
		mov ds, ax		;   data segment
		mov si, cs:[env_off]	;     and source.

		mov di, offset cs:[d_env_var]	; First character
		mov bl, cs:[di]		;   of environment variable.
;		Scan through environment for the end,
;		we need two null bytes.
loop:
		lodsb			; Get a byte.
		cmp al, 00h		; Is it null?
		jz zero1		; Yes. Got first null.
		jmp loop		; Keep looking.

;		Is the variable there?
zero1:
		lodsb			; Get a byte.
		cmp al, 00h		; Is it zero?
		jz write		; Yes.  Jump out.
		cmp al, bl		; Is it the env. variable?
		jnz loop		; No.  Jump out.

		dec si			; We are N+1.
		mov ax, ds		; Get the data segment
		mov es, ax		; and put it in extra segment.
		mov di, si		; This is our destination.

		mov dx, si		; Save the address of env. var.
		pop ds			; Restore the data segment.
		mov si, offset d_env_var ; Set the source.
		mov cx, s_env_var	; Length of variable.
		rep cmpsb		; Does it exist?
		jz update

		push ds			; Save the data segment.
		mov ax, es		; Get the extra segment.
		mov ds, ax		; and put it in data segment.
		mov si, di		; Setup the source.
		jmp loop

;		The environment variable is there.
;		Now update the environment variable.
update:
		push ds			; Save the data segment.
		mov ax, es		; Get the extra segment.
		mov ds, ax		; and put it in data segment.
		mov di, dx		; This is our destination.
		mov si, di		; Setup the source.
		
update1:
		lodsb			; Skip
		cmp al, 00h		;   over
		jnz update1		;     the variable.

		lodsb			; Look ahead to see
		dec si			;   if the next byte
		cmp al, 00h		;     is zero if not
		jz update3		;       compress the data.

update2:
		lodsb			; Compress
		stosb			;   the data
		cmp al, 00h		;     for a
		jnz update2		;       variable.

		lodsb			; Get a byte.
		dec si
		cmp al, 00h		; Is it zero?
		jnz update2		; No.  Get next.

update3:
		mov al, cs:[exit_code]
		cmp al, 05h
		jg update4
		mov cs:[exit_code], 05h	; Environment variable updated.
update4:
		mov cx, cs:[s_data]	; Length of data.
		mov dx, cs:[s_env_var]	; Length of name.
		sub cx, dx
		sub cx, 02h		; Minus 2 zeroes.
		cmp cx, 00h		; Is it just the name.
		jnz write1		; No. Jump out.

		mov cx, si		; End of old environment.
		mov dx, di		; Current position.
		sub cx, dx
		mov al, 00h		; Zero
		rep stosb		;  the rest.
		mov cs:[exit_code], 02h	; Environment variable erased.
		jmp nor_term

;		The environment variable isn't there.
;		Now move the data into place.
write:
		mov ax, ds		; Get the data segment
		mov es, ax		; and put it in extra segment.
		dec si			; We are N+1.
		mov di, si		; This is our destination.

write1:
		pop ds			; Restore the data segment.
		mov si, offset d_env_var ; Set the source.
		mov cx, s_data		; Set the length.
write2:
		mov ah, es:[di]		; Get destination byte.
write3:
		cmp ah, 00h		; Is it a zero?
		jnz not_zero		; No.  Jump out.
		lodsb			; Get byte.
		stosb			; Save byte.
		loop write2		; Loop.

;		Exit the program.
nor_term:
		mov al, cs:[exit_code]	; Set termination code.
		mov ah, 4ch		; Termination.
		int 21h

not_zero:
		mov ax, di		; Get the current offset.
		mov bx, env_off		; Get environment start.
		sub ax, bx		; Subtract the start of env.
		mov bx, env_siz		; Get environment size.
		cmp ax, bx		; Greater the environment size.
		jge reset		; Yes.  Jump.
		mov ah, 00h		; Make it zero.
		mov al, exit_code	; Get exit code.
		cmp al, 05h		; If update then do not
		jge write3		;   reduce the error code.
		mov exit_code, 04h	; Environment space over written.
		jmp write3		;   and redo.

reset:
		dec di
		dec di
		mov al, 00h		; Put zero back.
		stosb
		stosb
		mov exit_code, 10h	; Set code for out of env.,
		jmp nor_term		;   and terminate.

code ends
     end start

